from pydantic import BaseModel
from typing import Literal, TypeVar, Dict, Tuple, List, Hashable
import json
from abc import ABC, abstractmethod

# Pydantic model of a TM description

# You do not need to modify this file

class DTMCode(BaseModel):
    initial:  Hashable # Initial state
    accept:   List[Hashable] # Accepting states 
    reject:   List[Hashable] # Rejecting states
    transition : Dict[Hashable, Dict[str, Tuple[Hashable, str, Literal[-1, 0, 1]]]] # Transition table
    ## transition[state][read_letter] contains (next_state, new_letter, movement of the head in -1, 0, 1). 

    @property
    def states(self):
        states = set([self.initial])
        states.update(self.accept, self.reject, self.transition)
        for s, tr in self.transition.items():
            states.update((s for s,_,_ in tr.values()))
        return states

    def draw(self):
        """
        require graphviz installed (python module **and** software)
        """
        import graphviz
        dot = graphviz.Digraph()
        for a in self.states:
            shape = "circle"
            if a in self.accept:
                shape = 'doubleCircle'
            if a in self.reject:
                shape = "diamond"
            dot.node(a, shape=shape)

        for state_in, tr in self.transition.items():
            for letter, (state_out, letter_write, movement) in tr.items():
                dot.edge(state_in, state_out, ",".join(map(str, (letter, letter_write, movement))))
        return dot

    
class DTM(ABC):
    def __init__(self, code: DTMCode):
        self.code: DTMCode = code

    @abstractmethod
    def start(self, input_word):
        """ Initialize the TM on a given input word
        
        The first entry of the tape is symbol '^'
        Blank entries will be assumed to contain the symbol '_'
        Neither '^' and '_' can be part of the input word"""
        pass
        
    
    @abstractmethod
    def print(self):
        """Print the current configuration:  current state, content of the tape, position on the tape

        For example, it could print something like this:
        |^|a|a|b|a|b|
           ↑  start
        You can alternatively use colors to indicate the head position
        """
        pass

    @abstractmethod
    def next(self, print=False):        
        """ Execute one step of the transition
        
        - Return True if an accepted state is reached
        - Raise an exception:
            * if a rejecting state is reached
            * if the head moves out of the tape
        - Return False otherwise
        
        Print the configuration at each transition if print is set to True
        """
        pass

    @abstractmethod
    def exec(self, step=1000, print=False):
        """ Execute the machine until an accepting or rejecting state is reached, or `step` steps have been executed.
         
         Print the configuration at each transition if print is set to True
        """
        pass
